#include <iostream>
#include <iterator>
#include <numeric>
#include <ranges>

// struct-of-arrays
struct MyData {
    int dataPoint[8] = {};
    bool dataPointIsValid[8] = {};
};

// array-of-structs, and a container to put it in
struct MyDataPoint {
    int value = 0;
    bool isValid = false;
};
struct MyNicerData {
    MyDataPoint data[8];
};

// =========== Taking Sums ============

int sumOfDataPoints(const MyData& data)
{
    int sum = 0;
    for (int i = 0; i < 8; ++i) {
        sum += data.dataPoint[i];
    }
    return sum;
}

int sumOfNicerDataPoints(const MyNicerData& data)
{
    int sum = 0;
    for (int i = 0; i < 8; ++i) {
        sum += data.data[i].value;
    }
    return sum;
}

// === Taking Sums, Algorithmically ===

int stdSumOfDataPoints(const MyData& data)
{
    return std::accumulate(std::begin(data.dataPoint), std::end(data.dataPoint), 0);
}

int stdSumOfNicerDataPoints(const MyNicerData& data)
{
    return std::accumulate(std::begin(data.data), std::end(data.data), 0,
                           [](int v, const MyDataPoint& p) { return v + p.value; });
}

// ======= Algorithmic Take Two =======

int sumOfValidDataPoints(const MyData& data)
{
    int sum = 0;
    for (int i = 0; i < 8; ++i) {
        if (data.dataPointIsValid[i]) {
            sum += data.dataPoint[i];
        }
    }
    return sum;
}

int sumOfNicerValidDataPoints(const MyNicerData& data)
{
    int sum = 0;
    for (int i = 0; i < 8; ++i) {
        if (data.data[i].isValid) {
            sum += data.data[i].value;
        }
    }
    return sum;
}

int stdSumOfValidNicerDataPoints(const MyNicerData& data)
{
    return std::accumulate(std::begin(data.data), std::end(data.data), 0,
                           [](int v, const MyDataPoint& p) { return p.isValid ? v + p.value : v; });
}

#if 0
// This cannot compile: we do no have an index to apply to
// the members of MyData.
int stdSumOfValidDataPoints(const MyData& data)
{
    return std::accumulate(std::begin(data.dataPoint), std::end(data.dataPoint), 0,
                           /* How to get data.dataPointIsValid here? */);
}
#endif

// This does not compile on Clang (up to at least clang 15) because of limitations
// in the standard ranges library implementation.
int iotaSumOfValidDataPoints(const MyData& data)
{
    auto r = std::ranges::iota_view(std::size_t(0), std::size(data.dataPoint));
    return std::accumulate(r.begin(), r.end(), 0,
                           [&data](int v, std::size_t index) { return data.dataPointIsValid[index] ? v + data.dataPoint[index] : v; });
}

int main()
{
    // Some arbitrary data to work on
    MyNicerData data_nice = {
        { { 13, true },
          { 2, true },
          { 4, false },
          { 7, true },
          { 113, false },
          { 1, false },
          { 11, false },
          { 57, true } }
    };
    MyData data_ugly;
    for (int i = 0; i < 8; ++i) {
        data_ugly.dataPoint[i] = data_nice.data[i].value;
        data_ugly.dataPointIsValid[i] = data_nice.data[i].isValid;
    }

    std::cout << "== MyData (struct-of-arrays) ==\n";
    std::cout << "    sum=" << sumOfValidDataPoints(data_ugly) << '\n';
    std::cout << " stdsum=" << stdSumOfDataPoints(data_ugly) << '\n';
    std::cout << "iotasum=" << iotaSumOfValidDataPoints(data_ugly) << '\n';

    std::cout << "== MyNicerData (array-of-structs) ==\n";
    std::cout << "    sum=" << sumOfNicerValidDataPoints(data_nice) << '\n';
    std::cout << " stdsum=" << stdSumOfNicerDataPoints(data_nice) << '\n';
    std::cout << "vstdsum=" << stdSumOfValidNicerDataPoints(data_nice) << '\n';
}
